% 2004/10/26. Version of graphs using accumulators. 

:- ensure_loaded( '../src/init_lib' ).
:- ensure_loaded( library(gnucomfile) ).			% /2.
:- ensure_loaded( library(system) ).				% exec/3, delete_file/1.
:- ensure_loaded( library(mktemp_in_dir) ).				% /3.
:- ensure_loaded( library(flatoms) ).					% /2.
:- ensure_loaded( library(kv_transpose) ).				% /3.

:- multifile( exec_extension/2 ).
:- dynamic( exec_extension/2 ).

exec_extension( agraphs, BareFile ) :-
	atom_concat( BareFile, '.stays', Stays ),
	atom_concat( BareFile, '.visits', Visits ),
	open( Stays, write, StS ),
	open( BareFile, read, BrS ),
	rec_model_sv_counts( BrS, StS, Visits ),
	close( StS ), 
	close( BrS ),
	mktemp_in_dir( gnuplotcoms_XXXXXX, '.', GnuComsFile ),
	flatoms( ['gnuplot -persist \"',GnuComsFile,'\"'], Gplot1 ),
	gnucomfile_w_lines( GnuComsFile, Stays ),
	exec( Gplot1, [null,null,null], _Pid1 ),
	sleep( 3 ),
	gnucomfile_w_lines( GnuComsFile, Visits ),
	exec( Gplot1, [null,null,null], _Pid2 ),
	sleep( 2 ),
	delete_file( GnuComsFile ).

rec_model_sv_counts( In, OutSty, VisFile ) :-
	next_chain_model( In, First ),
	next_chain_model( In, Second  ),
	rec_model_sv_counts_1( Second, First, In, OutSty, 1, [], Vis ),
	kv_transpose( Vis, Tsp ),
	keysort( Tsp, SrtTsp ),
	tell( VisFile ),
	write_visited_gnuplot_lines( SrtTsp ),
	told.

rec_model_sv_counts_1( end_of_file, Curr, _In, OutSty, CurrCnt, AccVis, Vis ) :-
	!,
	write( OutSty, CurrCnt ), nl( OutSty ),
	update_vis_accumulator( AccVis, Curr, CurrCnt, Vis ).
rec_model_sv_counts_1( Curr, Curr, In, OutSty, CurrCnt, AccVis, Vis ) :-
	!,
	next_chain_model( In, Nxt ),
	NxCCnt is CurrCnt + 1,
	rec_model_sv_counts_1( Nxt, Curr, In, OutSty, NxCCnt, AccVis, Vis ).
rec_model_sv_counts_1( Nxt, Curr, In, OutSty, CurrCnt, AccVis, Vis  ) :-
	write( OutSty, CurrCnt ), nl( OutSty ),
	update_vis_accumulator( AccVis, Curr, CurrCnt, NxAccVis ),
	next_chain_model( In, Ahead ),
	rec_model_sv_counts_1( Ahead, Nxt, In, OutSty, 1, NxAccVis, Vis  ).

next_chain_model( In, Next ) :-
	read( In, Term ),
	next_chain_model_1( Term, In, Next ).

next_chain_model_1( end_of_file, _In, end_of_file ) :- !.
next_chain_model_1( c(Mod), _In, Mod ) :- !.
next_chain_model_1( _Term, In, Mod ) :-
	read( In, NxtTerm ),
	next_chain_model_1( NxtTerm, In, Mod ).

update_vis_accumulator( [HM-HC|T], Model, Count, [HM-HC|TUpd] ) :-
	HM @< Model,
	!, 
	update_vis_accumulator( T, Model, Count, TUpd ).
update_vis_accumulator( [HM-HC|T], Model, Count, Upd ) :-
	HM == Model,
	!,
	UpdC is HC + Count,
	Upd = [HM-UpdC|T].
update_vis_accumulator( AccTail, Model, Count, [Model-Count|AccTail] ).

write_visited_gnuplot_lines( [] ).
write_visited_gnuplot_lines( [Tms-Mod|T] ) :-
	write( Tms ), write( '   #   ' ), writeq( Mod ), write( '.' ), nl,
	write_visited_gnuplot_lines( T ).
