:- ensure_loaded( '../src/init_lib' ).
:- ensure_loaded( library(lists) ).		% memberchk/2.
:- ensure_loaded( library(requires) ).

:- requires( opts_have_help/1 ).
:- requires( opts_select_infiles/3 ).
:- requires( read_terms/2 ).
:- requires( write_terms/2 ).
:- requires( intervals/4 ).
:- requires( kv_compose_fixed_v/3 ).

predicted_diff_bins( Opts ) :-
	( opts_have_help(Opts) ->
		predicted_diff_bins_help_message
		;
		opts_select_infiles( Opts, Files, FlOpts ),
		( Files = [First,Second] ->
			predicted_diff_bins_1( FlOpts, First, Second )
			;
			predicted_diff_bins_help_message
		)
	).
	
predicted_diff_bins_1( Opts, First, Second ) :-
	( memberchk(out(Out),Opts) ->
		( memberchk(max(Max),Opts) -> true; Max is 1 ),
		Min is - Max,
		( memberchk(bin(Bin),Opts) -> true; Bin is 0.05 ),
		intervals( Min, Max, Bin, [_|ULims] ), % upper limit for each bin
		kv_compose_fixed_v( ULims, 0, EmptyBins ),
		read_terms( First, FstPctions ),
		read_terms( Second, SecPctions ),
		prediction_lists_binned( FstPctions, SecPctions, EmptyBins, DiffBins ),
		kv_write_both_with_sep_space( Out, DiffBins )
		;
		predicted_diff_bins_help_message
	).

prediction_lists_binned( [], [], Bins, Bins ).
prediction_lists_binned( [[_-H1|_]|T1], [[_-H2|_]|T2], [LimH-SeenH|AccT], Bins ) :-
	Diff is H1 - H2,
	add_to_bins( AccT, LimH, SeenH, Diff, NxBins ),
	prediction_lists_binned( T1, T2, NxBins, Bins ).

add_to_bins( [], LimH, SeenH, _Diff, Bins ) :-
	NxSeenH is SeenH + 1,
	Bins = [LimH-NxSeenH].
add_to_bins( [H|T], Lim, Seen, Diff, Bins ) :-
	Lim < Diff,
	!,
	H = LimH-SeenH,
	Bins = [Lim-Seen|TBins],
	add_to_bins( T, LimH, SeenH, Diff, TBins ).
add_to_bins( [H|T], Lim, Seen, _Diff, Bins ) :-
	NxSeen is Seen + 1,
	Bins = [Lim-NxSeen,H|T].

kv_write_both_with_sep_space( File, KVs ) :-
	open( File, write, Out ),
	kv_write_both_with_sep_space_1( KVs, Out ),
	close( Out ).

kv_write_both_with_sep_space_1( [], _Out ).
kv_write_both_with_sep_space_1( [K-V|T], Out ) :-
	write( Out, K ), write( Out, ' ' ), write( Out, V ), nl( Out ),
	kv_write_both_with_sep_space_1( T, Out ).

predicted_diff_bins_help_message :-
	write( 'predicted_diff_bins/1: See sources for usage.' ), nl, abort.
