
/* 
     cohese and visualise a number of .ecounts files 

     upsh ~/islp/mcmcms/bns/auxil/edge_counts_viz tr_uc_rm_binari_freq_vanSteensel_bin_i500K__s221.ecounts 0.9

     upsh ~/islp/mcmcms/bns/auxil/edge_counts_viz *.ecounts ids=/usr/local/users/nicos/data/vanSteensel/work/SuppData1.txt r_out=file.r 0.9
     */

:- requires( skim/3 ).

:- ensure_loaded( edge_counts_select ).
% :- ensure_loaded( '~/pl/programs/r_session/r_session_swi.pl').
:- ensure_loaded( '~/pl/programs/r_session/r_session_yap.pl').

:- ensure_loaded( library(apply) ).
:- ensure_loaded( library(ordsets) ).
:- ensure_loaded( library(readutil) ).

edge_counts_viz( Inps ) :-
     append( Opts, [Cutoff], Inps ),
     partition( opt_name(psf), Opts, Psfxs, NpsfOpts ),
     ( select(ids(IdsFile),NpsfOpts,IdsOpts) -> 
          true
          ;
          IdsOpts = Opts,
          IdsFile = null__
     ),
     ( select(clrs(ClrsF),IdsOpts,IFiles) ->
          ensure_loaded(ClrsF)
          ;
          throw( you_must_give_a_colour_file )
     ),
     ( select(r_out(RFile),IFiles,Files) -> 
          true
          ;
          Files = IFiles,
          atomic_list_concat( [coff_,Cutoff,'.r'], RFile )
          %% older approach : RFile = null__
     ),
     ( number(Cutoff) -> 
          true
          ;
          throw( not_a_number(Cutoff) )
     ),
     maplist( file_to_edge_count_quartets, Files, Qrts ),
     length( Qrts, Length ), 
     write( nof_files(Length) ), nl,
     Qrts = [Q|_Qs],
     edge_counts_quarters_cohese( Q, Qrts, Length, Trips, NNodes ),
     list_to_ord_set( NNodes, Nodes ),
     length( Nodes, LNodes ),
     edge_trips_select( Trips, Cutoff, Select, Combined ),
     ( IdsFile == null__ -> 
          findall( Id, between(0,LNodes,Id), Ids )
          ;
          first_csv_line( IdsFile, [_ID|Ids] )
     ),
     findall( S, (member(S,Select), S=A-B-C, 
                    % Ac is A - 1, Bc is B - 1,
                    nth1( A, Ids, Ath ),
                    nth1( B, Ids, Bth ),
                    write(Ath), write( ' ' ), write(Bth), write( ' ' ),
                    write(C), nl), _ ),
     write( combined ), nl,
     findall( S, (member(S,Combined), S=A-B-C, 
                    % Ac is A - 1, Bc is B - 1,
                    nth1( A, Ids, Ath ),
                    nth1( B, Ids, Bth ),
                    write(Ath), write( ' ' ), write(Bth), write( ' ' ),
                    write(C), nl), _ ),
     write( no_weight ), nl,
     findall( S, (member(S,Combined), S=A-B-C, 
                    % Ac is A - 1, Bc is B - 1,
                    nth1( A, Ids, Ath ),
                    nth1( B, Ids, Bth ),
                    write(Ath), write( ' ' ), write(Bth), write( ' ' ), nl), _ ),
     length( Select, Selected ),
     write( nof_selected(Selected) ), nl,
     edge_count_selected_viz( Select, LNodes, IdsFile, RFile, Psfxs ).

edge_counts_quarters_cohese( [], _, _Length, [], [] ).
edge_counts_quarters_cohese( [_|T], Qrts, Length, [X-Y-Avg|TTrips], Nds ) :-
     skim( Qrts, Skim, Scum ),
     Skim = [X-Y-Tms-Tot|TSkim],
     average_quartets( TSkim, X, Y, Tms, Tot, 1, Avg, QLength ),
     ( Length \== QLength -> 
          throw( length_mismatch(Length,QLength) )
          ;
          true
     ),
     Nds = [X,Y|TNds],
     edge_counts_quarters_cohese( T, Scum, Length, TTrips, TNds ).

average_quartets( [], _X, _Y, Tms, Tot, L, Avg, L ) :-
     Avg is Tms / Tot.
average_quartets( [X-Y-Tms-Tot|T], X, Y, AccTms, AccTot, I, Avg, Length ) :-
     NxtTms is AccTms + Tms,
     NxtTot is AccTot + Tot,
     NxtI is I + 1,
     average_quartets( T, X, Y, NxtTms, NxtTot, NxtI, Avg, Length ).

edge_counts_viz_1( File, Cutoff ) :-
     edge_counts_select( File, Cutoff, Selects ),
     edge_count_selected_viz( Selects, null__, null__, [] ).

edge_count_selected_viz( Selects, N, IdsFile, RFile, Psfxs ) :-
     findall( [X,Y], (member(A-B-_W,Selects),X is A - 1, Y is B - 1), NEdges ),
     flatten( NEdges, Edges ),
     r_open,
     r_lib( igraph ),
     r_in( edges <- Edges ),
     ( RFile == null__ -> true; record_to_r_file(RFile,Edges) ),
     ( IdsFile == null__ -> 
          TkPlot = tkplot( graph(edges,n=N) )
          ;
          first_csv_line( IdsFile, [_ID|Entries] ),
          r_atoms_to_atomic_concat( Entries, 'c("', Vlabels ),
          write( vlabels(Vlabels) ), nl,
          r_in( vlabels <- Vlabels ),
          findall( Cc, (member(Ide,Entries),once(colour_code(Ide,Cc))), ColourCodes ),
          edge_colouring( Edges, Entries, EcolList, Bn80Es, Bn80Col ),
          r_atoms_to_atomic_concat( Bn80Es, 'c("', Bn80Edges ),
          r_in( bn80edges <- Bn80Edges ),
          r_atoms_to_atomic_concat( ColourCodes, 'c("', Vcolours ),
          r_atoms_to_atomic_concat( EcolList, 'c("', Ecolours ),
          r_atoms_to_atomic_concat( Bn80Col, 'c("', Bn80colours ),
          r_in( vcolours <- Vcolours ),
          r_in( ecolours <- Ecolours ),
          r_in( bn80colours <- Bn80colours ),
          add_to_r_file( RFile, vlabels <- Vlabels ),
          add_to_r_file( RFile, vcolours <- Vcolours ),
          add_to_r_file( RFile, ecolours <- Ecolours ),
          add_to_r_file( RFile, bn80edges <- Bn80Edges ),
          add_to_r_file( RFile, bn80colours <- Bn80colours ),
          TkPlot = tkplot(graph(edges,n=N),layout='layout.fruchterman.reingold','vertex.label'=vlabels,'vertex.color'=vcolours,'edge.color'=ecolours,'canvas.width'=1000,'canvas.height'=800) 
          
     ),
     add_to_r_file( RFile, TkPlot ),
     Bn80TkPlot = tkplot(graph(bn80edges,n=N),layout='layout.fruchterman.reingold','vertex.label'=vlabels,'vertex.color'=vcolours,'edge.color'=bn80colours,'canvas.width'=600,'canvas.height'=600),
     add_to_r_file( RFile, Bn80TkPlot ),
     findall( Psf, (member(psf(Psf),Psfxs),add_to_r_file(RFile,Psf)), _AllPsfxs ),
     r_in( TkPlot ),
     sleep( 1000 ),
     r_close.

add_to_r_file( null__, _This ) :- !.
add_to_r_file( File, This ) :-
     open( File, append, Out ),
     write( Out, This ),
     nl( Out ),
     close( Out ).

record_to_r_file( RFile, Edges ) :-
     open( RFile, write, Out ),
     write( Out, library(igraph) ), nl( Out ),
     write( Out, edges ), 
     write( Out, ' <- c(' ), 
     record_r_vector( Edges, Out ),
     nl( Out ),
     close( Out ).

record_r_vector( [], _Out ).
record_r_vector( [H|T], Out ) :-
     write( Out, H ), 
     ( T == [] ->
          write( Out, ')' )
          ;
          write( Out, ',' )
     ),
     record_r_vector( T, Out ).

r_atoms_to_atomic_concat( [], Left, Vlabels ) :-
     atom_concat( Left, '")', Vlabels ).
r_atoms_to_atomic_concat( [H|T], Left, Vlabels ) :-
     atomic_concat( Left, H, LeftH ), 
     ( T == [] -> 
          Next = LeftH 
          ;
          atom_concat( LeftH, '","', Next )
     ),
     r_atoms_to_atomic_concat( T, Next, Vlabels ).

file_to_edge_count_quartets( File, Qrts ) :-
     read_terms( File, ECs ),
     findall( X-Y-Tms-Tot, ( 
                         member( ec(X,Tot,Ys), ECs ),
                         nth1( Y, Ys, Tms ),
                         X \== Y
                     ),
                         Qrts ).

first_csv_line( File, Entries ) :-
     open( File, read, In ),
     read_line_to_codes( In, Line ),
     close( In ),
     first_csv_line_entries( Line, Entries ).

first_csv_line_entries( Line, Entries ) :-
     append( ThisCs, [0',|T], Line ),
     !,
     atom_codes( This, ThisCs ),
     Entries = [This|TEntries],
     first_csv_line_entries( T, TEntries ).
first_csv_line_entries( Line, [Entry] ) :-
     atom_codes( Entry, Line ).

edge_colouring( [], Entries, [], Bn80Es, ColBn80 ) :-
     findall( [X,Y], 
                    ( missing_edge(Xth,Yth),
                      nth0( X, Entries, Xth ),
                      nth0( Y, Entries, Yth )
                    ),
                         XYs ),
     flatten( XYs, Bn80Es ),
     length( Bn80Es, NodesInEdges ),
     L is integer(NodesInEdges / 2),
     findall( royalblue, between(1,L,_), ColBn80 ).
                              
edge_colouring( [X,Y|T], Entries, [Col|TEc], Bn80Es, ColBn80 ) :-
     % X1 is X + 1,
     % Y1 is Y + 1,
     % nth0( X1, Entries, Xth ),
     % nth0( Y1, Entries, Yth ),
     nth0( X, Entries, Xth ),
     nth0( Y, Entries, Yth ),
     ( Xth == 'PROD' -> trace; true ),
     ( (new_edge(Xth,Yth);new_edge(Yth,Xth)) ->
          write( new(Xth,Yth) ), nl,
          Col = darksalmon,
          Bn80Es = TBn80Es,
          ColBn80 = TCBn80
          ;
          write( old(Xth,Yth) ), nl,
          Bn80Es = [X,Y|TBn80Es],
          ColBn80 = [grey|TCBn80],
          Col = grey
     ),
     edge_colouring( T, Entries, TEc, TBn80Es, TCBn80 ).

opt_name( Name, Elem ) :-
     functor( Elem, Name, _ ).
