% 7 full dots for 400000000 over 8 nodals. one dot takes more than a day
% when my computer is used each dot takes just over two days for 
% a new dot (6 -> 7 ).
% [.0] .........

:- ensure_loaded( library(lists) ).		% memberchk/2.
:- ensure_loaded( library(ugraphs) ).		% add_edges/3.
:- ensure_loaded( library(ordsets) ).		% rpdag/2.

:- ensure_loaded( '../../src/init_lib' ).
:- ensure_loaded( 'src/rpdag_to_bn' ).		%
:- ensure_loaded( '../auxil/disp_bn' ).

test( V ) :- rpdag( [1,2,3], V ).

doubles :-
	findall(R, rpdag([1,2,3,4,5,6],R), Rs ),
	length( Rs, Length ), 
	write( all_found_length(Length) ), nl,
	doubles_1( Rs, 1, [] ).

doubles_1( [], _N, Unique ) :-
	length( Unique, Length ), 
	write( uniques_length(Length) ), nl.
doubles_1( [H|T], N, Acc ) :-
	( memberchk(H,T) -> 
		write( duplicate(N,H) ), nl,
		NxAcc = Acc
		;
		NxAcc = [H|Acc]
	),
	NxN is N + 1,
	doubles_1( T, NxN, NxAcc ).

test :- count_connect.

variant( H, H ).
count_connect :-
	ensure_loaded( library('list_frequency') ),
	ensure_loaded( library('graph/symmetric_reduction') ),
	findall( R-L, 
		(rpdag([1,2,3,4],R),
		 count_connections( R, L )
		 ),
		RLs ),
	kv_decompose( RLs, Rs, Ls ),
	list_frequency( Ls, Frq ),
	list_connected_as( Rs, (2,3) ),
	write( Frq ), nl.
count_connections( rpdag(_Ps,Dir,Und), L-(NofUnd,NofDir) ) :-
	edges( Dir, Edges ),
	length( Edges, NofDir ),
	symmetric_reduction( Und, RedUnd ),
	edges( RedUnd, RedUndEdges ),
	length( RedUndEdges, NofUnd ),
	L is NofDir + NofUnd.

list_connected_as( [], _Pair ).
list_connected_as( [H|T], Pair ) :-
	count_connections( H, _-HPair ),
	( Pair \= HPair -> 
		true
		; 
		write( H ), nl,
		disp_rpdag( H )
	),
	list_connected_as( T, Pair ).

count( Every ) :- 
	current_output( Was ), bb_put( output_was, Was ),
	open_null_stream( Null ), 
	% open( all_four_node_rpdags, write, Out ),
	set_output( Null ),
	bb_put( i, 0 ),
	Nodes = [1,2,3,4,5,6,7,8],
	rpdag( Nodes, _ ),
	% rpdag( Nodes, rpdag(_Ps,Dir,Und) ),
	% edges( Dir, DirEdges ), edges( Und, UndEdges ),
	% length( DirEdges, E ), length( UndEdges, U ),
	% count( 400000000 ).
	% write( user_error, E*U ), nl( user_error ),
	% ( (U=:=1,E=:=4) -> 
		% write( user_error, DirEdges+UndEdges ), nl( user_error) ; true ),
	% write( Out, DirEdges+UndEdges ), nl( Out ),
	bb_get( i, I ),
	( 0 =:= I mod Every -> 
		write( user_error, '.' )
		;
		true
	),
	NxI is I + 1,
	bb_put( i, NxI ),
	fail.
count( _ ) :-
	bb_get( i, I ),
	bb_get( output_was, Was ),
	set_output( Was ),
	write( found(I) ), nl.

viz :- 
	bb_put( i, 1 ),
	bb_put( d, 1 ),
	ensure_loaded( disp_rpdags ),
	Nodes = [1,2,3,4],
	rpdag( Nodes, rpdag(Pairs,Dir,Und) ),
	bb_get( i, I ),
	( bb_get( d, I ) ->
		% ( I = 27 -> trace ; true ),
		edges( Dir, DirEdges ),
		add_edges( Und, DirEdges, Full ),
		transpose( Full, BnLike ),
		add_vertices( BnLike, Nodes, BnFull ),
		disp_bn( BnFull ),
		rpdag_to_bn( rpdag(Pairs,Dir,Und), ExtBn ), 
		disp_bn( ExtBn ),
		write( rpdag(Pairs,Dir,Und) ), nl,
		write( i(I) ), write( ' : ' ), read( Inc ),
		( number(Inc) -> 
			true ; Inc = 1 ),
		D is I + Inc,
		bb_put( d, D )
		;
		true
	),
	NxI is I + 1,
	bb_put( i, NxI ),
	fail.

disp_rpdag( rpdag(_,Dir,Und) ) :-
	edges( Dir, DirEdges ),
	add_edges( Und, DirEdges, Full ),
	transpose( Full, BnLike ),
	add_vertices( BnLike, [1,2,3,4], BnFull ),
	disp_bn( BnFull ).

rpdag( [], rpdag([],[],[]) ).
rpdag( [H|T], Dag ) :-
	rpdag( T, DagT ),
	null_node_info( Hninf ),
	DagT = rpdag(InPairs,Dir,Und),
	% ( (Und = [2-[3],3-[2,4],4-[3]], Dir = []) -> bb_put(trace,on) ; true ),
	ord_add_element( InPairs, H-Hninf, NxPairs ),
	add_vertices( Dir, [H], HDir ),
	add_vertices( Und, [H], HUnd ),
	DagH = rpdag(NxPairs,HDir,HUnd),
	% ( (H =:= 1,Und=[]) -> trace ; true ),
	connect( T, DagH, H, Hninf, Dag ).

null_node_info( ([]/0,[]/0,[]/0)).

connect( [], Dag, _X, _Xninf, Dag ).
connect( [Y|T], InDag, X, Xninf, Dag ) :-
	InDag = rpdag(Pairs,_,_),
	memberchk( Y-Yninf, Pairs ),
	connect_node( Y, Yninf, X, Xninf, InDag, NxXnInf, NxDag ),
	connect( T, NxDag, X, NxXnInf, Dag ).

% 3/4 :: % K / # nodes in G 
connect_node( Y, _Yninf, X, Xninf, Dag, Xninf, Dag ) :-
	write( connection_none(X,Y) ), nl.
% 1/4 ::
connect_node( Y, Yninf, X, Xninf, InDag, OutXninf, NxDag ) :-
	% X not in AdY, \+ memberchk( X, AdY )
	connect_node_1( Y, Yninf, X, Xninf, InDag, OutXninf, NxDag ),
	write( inxninf(Xninf) ), nl,
	write( outxninf(OutXninf) ), nl.

% 1/5 :: (each)
% A.arc(x,y) % check order is correct
connect_node_1( Y, Yninf, X, Xninf, InDag, OutXninf, OutDag ) :-
	write( attempting_to_add_arc_x_y(X,Y) ), nl,
	Yninf = (_PaY/Py,_ChY/Cy,_NeY/Ny),
	Xninf = (PaX/Px,ChX/Cx,NeX/Nx),
	(Px > 0 -> 
		true
		;
		Py > 0
	),
	% InDag = rpdag(_InPairs,InDir,_InUnd),
	add_arc_if( Px, Ny, Cy, X, Y, InDag, OutDag),
	ord_add_element( ChX, Y, NwChX ),
	NwCx is Cx + 1,
	OutXninf = (PaX/Px,NwChX/NwCx,NeX/Nx),
	write( done_adding_arc_x_y(X,Y) ), nl.
% A.arc(y,x)
connect_node_1( Y, Yninf, X, Xninf, InDag, OutXninf, OutDag ) :-
	write( attempting_to_add_arc_y_x(Y,X) ), nl,
	Yninf = (_PaY/Py,_ChY/_Cy,_NeY/_Ny),
	Xninf = (PaX/Px,ChX/Cx,NeX/Nx),
	(Px > 0 ->
		true
		;
		Py > 0
	),
	% InDag = rpdag(_InPairs,InDir,_InUnd),
	add_arc_if( Py, Nx, Cx, Y, X, InDag, OutDag ),
	ord_add_element( PaX, Y, NwPaX ),
	NwPx is Px + 1,
	OutXninf = (NwPaX/NwPx,ChX/Cx,NeX/Nx),
	write( done_adding_arc_y_x(Y,X) ), nl.
% A.link(x,y)
connect_node_1( Y, Yninf, X, Xninf, InDag, OutXninf, OutDag ) :-
	% ((bb_get(trace,on),X=1,Y=2)->trace;true),
	write( attempting_to_add_link(X,Y) ), nl,
	Yninf = (_PaY/Py,_ChY/_Cy,_NeY/Ny),
	Xninf = (_PaX/Px,_ChX/_Cx,_NeX/Nx),
	Py = 0, Px = 0,
	add_link_if( Nx, X, Ny, Y, InDag, OutDag ),
	% ord_add_element( NeX, Y, NwNeX ),
	% ord_add_element( AdX, Y, NwAdX ),
	% NeX is 
	OutDag = rpdag(OutPairs,_,_),
	memberchk( X-OutXninf, OutPairs ),
	%% ( (X=1,Y=2) -> bb_put(trace,on) ; true ),
	write( done_adding_link(X,Y) ), nl.
% A.hh(x,y)
connect_node_1( Y, Yninf, X, Xninf, InDag, OutXninf, OutDag ) :-
	% (bb_get(trace,on)->trace;true),
	write( attempting_to_add_hh(X,Y) ), nl,
	Yninf = (_PaY/Py,_ChY/_Cy,NeY/Ny),
	Xninf = (_PaX/Px,_ChX/_Cx,_NeX/Nx),
	Py = 0,
	Ny > 0,
	% pmember ?
	member( Z, NeY ),
	%%( (X=1,Y=3,Z=4,bb_get(trace,on)) -> trace ; true ),
	% rpdag([1-([]/0,[]/0,[2]/1),2-([]/0,[]/0,[1,3]/2),3-([]/0,[]/0,[2,4]/2),4-([]/0,[]/0,[3]/1)],[1-[],2-[],3-[],4-[]],[1-[2],2-[1,3],3-[2,4],4-[3]])
	add_hh_if( Px, Nx, X, Ny, Y, Z, Xninf, InDag, OutDag ),
	OutDag = rpdag(OutPairs,_,_),
	memberchk( X-OutXninf, OutPairs ),
	write( done_adding_hh(X,Y,Z) ), nl.
% A.hh(y,x)
connect_node_1( X, Xninf, Y, Yninf, InDag, OutXninf, OutDag ) :-
	write( attempting_to_add_hh_y_x(X,Y) ), nl,
	Yninf = (_PaY/Py,_ChY/_Cy,NeY/Ny),
	Xninf = (_PaX/Px,_ChX/_Cx,_NeX/Nx),
	Py = 0,
	Ny > 0,
	% pmember ?
	member( Z, NeY ),
	add_hh_if( Px, Nx, X, Ny, Y, Z, Xninf, InDag, OutDag ),
	OutDag = rpdag(OutPairs,_,_),
	memberchk( Y-OutXninf, OutPairs ),
	write( done_adding_hh_y_x(X,Y,Z) ), nl.

% A.hh(x,y) starts
add_hh_if( Px, Nx, X, Ny, Y, Z, Xninf, InDag, OutDag ) :-
	Ny >= 2,
	(Px > 0; Nx > 0),
	!,
	% ( (X=1,Y=3) -> trace ; true ),
	write( committed_hh(ny(Ny),px(Px),nx(Nx)) ), nl,
	InDag = rpdag(_Pairs,InDir,_InUnd),
	reachable( Y, InDir, YReaches ),
	\+ memberchk( X, YReaches ),
	\+ memberchk( Z, YReaches ),
	add_hh_post_if( Ny, X, Y, Z, Xninf, InDag, OutDag ).
add_hh_if( _Px, _Nx, X, Ny, Y, Z, Xninf, InDag, OutDag ) :-
	add_hh_post_if( Ny, X, Y, Z, Xninf, InDag, OutDag ).

add_hh_post_if( InNy, X, Y, Z, _TheXninf, InDag, OutDag ) :-
	InDag = rpdag(Pairs,InDir,InUnd),
	once( select(Y-Yninf,Pairs,NoYPairs) ),
	once( select(X-Xninf,NoYPairs,NoXPairs) ),
	once( select(Z-Zninf,NoXPairs,NoZPairs) ),
	% remove_undirected( InUnd, Y-Z, MdUnd ),
	del_edges( InUnd, [Y-Z,Z-Y], MdUnd ),
	add_edges( InDir, [X-Y,Z-Y], MdDir ),
	Yninf = (InPaY/InPy,ChY/Cy,InNeY/InNy),
	Xninf = (PaX/Px,InChX/InCx,NeX/Nx),
	Zninf = (PaZ/Pz,InChZ/InCz,InNeZ/InNz),
	ord_add_element( InChX, Y, ChX ), Cx is InCx + 1,
	ord_add_element( InChZ, Y, ChZ ), Cz is InCz + 1,
	ord_del_element( InNeY, Z, NeY ), Ny is InNy - 1,
	ord_add_element( InPaY, X, MdPaY ),
	ord_add_element( MdPaY, Z, PaY ), Py is InPy + 2,
	ord_del_element( InNeZ, Y, NeZ ), Nz is InNz - 1,
	UpYninf = (PaY/Py,ChY/Cy,NeY/Ny),
	UpXninf = (PaX/Px,ChX/Cx,NeX/Nx),
	UpZninf = (PaZ/Pz,ChZ/Cz,NeZ/Nz),
	ord_add_element( NoZPairs, Y-UpYninf, UpYPairs ),
	ord_add_element( UpYPairs, X-UpXninf, UpXPairs ),
	ord_add_element( UpXPairs, Z-UpZninf, UpZPairs ),
	( InNy >= 2 ->
		complete( [Y], UpZPairs, MdDir, MdUnd, CompYPairs, CompYDir, CompYUnd )
		;
		CompYPairs = UpZPairs, CompYDir = MdDir, CompYUnd = MdUnd
	),
	top_sort( CompYDir, _Sorted ),
	OutDag = rpdag(CompYPairs,CompYDir,CompYUnd).
% add_hh_post_if( _Ny, X, Y, Z, Xninf, InDag, InDag ) :-

% A.hh(x,y) ends

% A.link(x,y) starts
add_link_if( Nx, X, Ny, Y, InDag, OutDag ) :-
	Ny > 0, Nx > 0, !,
	write( committed_link(ny(Ny),nx(Nx)) ), nl,
	InDag = rpdag(InPairs,InDir,InUnd),
	reachable( X, InUnd, Reaches ),
	\+ memberchk( Y, Reaches ),
	add_link( InPairs, InDir, InUnd, X, Y, OutDag ).

add_link_if( _Nx, X, _Ny, Y, InDag, OutDag ) :-
	InDag = rpdag(InPairs,InDir,InUnd),
	add_link( InPairs, InDir, InUnd, X, Y, OutDag ).

add_link( InPairs, InDir, InUnd, X, Y, OutDag ) :-
	once( select( Y-Yninf, InPairs, NoYPairs ) ),
	% ord_del_element( InPairs, Y-Yninf, NoYPairs ),
	Yninf = (PaY/Py,ChY/Cy,NeY/Ny),
	ord_add_element( NeY, X, NwNeY ),
	NwNy is Ny + 1,
	NwYninf = (PaY/Py,ChY/Cy,NwNeY/NwNy),
	ord_add_element( NoYPairs, Y-NwYninf, NwYPairs ),
	once( select( X-Xninf, NwYPairs, NoXPairs) ),
	Xninf = (PaX/Px,ChX/Cx,NeX/Nx),
	ord_add_element( NeX, Y, NwNeX ),
	NwNx is Nx + 1,
	NwXninf = (PaX/Px,ChX/Cx,NwNeX/NwNx),
	ord_add_element( NoXPairs, X-NwXninf, NwPairs ),
	add_edges( InUnd, [X-Y,Y-X], NwUnd ),
	OutDag = rpdag(NwPairs,InDir,NwUnd).

% A.arc(x,y) starts
add_arc_if( Px, Ny, Cy, X, Y, InDag, OutDag ) :-
	Px > 0,
	( Ny > 0 ; Cy > 0),
	!,
	write( committed_arc(px(Px),ny(Ny)) ), nl,
	InDag = rpdag(InPairs,InDir,InUnd),
	reachable( X, InDir, Xreaches ),
	\+ memberchk( Y, Xreaches ),
	in_pairs_add_arc( InPairs, X, Y, XYPairs ),
	add_edges( InDir, [X-Y], XYDir ),
	( Ny > 0 ->
		complete( [Y], XYPairs, XYDir, InUnd, CompYPairs, CompYDir, CompYUnd )
		;
		CompYPairs = XYPairs, CompYDir = XYDir, CompYUnd = InUnd
	),
	top_sort( CompYDir, _Sorted ),
	% write( top_sort( CompYDir, Sorted ) ), nl,
	OutDag = rpdag(CompYPairs,CompYDir,CompYUnd).
	% needed ?
	% also maybe we could do better, since cycle(s)
	% must contain X->Y or newly converted link to arc. (conplete/n)
	% actually, there a will be a set of paths "generated" by complete/n
	% only need to consider cycles that may contain those,
	% from that, just check that there is no directed cycle form
	% each endpoint to X
add_arc_if( _Px, Ny, _Cy, X, Y, InDag, OutDag ) :-
	InDag = rpdag(InPairs,InDir,InUnd),
	in_pairs_add_arc( InPairs, X, Y, XYPairs ),
	add_edges( InDir, [X-Y], XYDir ),
	% ( (Px > 0, Ny > 0) -> 
	( (Ny > 0) -> 
		complete( [Y], XYPairs, XYDir, InUnd, CompYPairs, CompYDir, CompYUnd )
		;
		CompYPairs = XYPairs, CompYDir = XYDir, CompYUnd = InUnd
	),
	top_sort( CompYDir, Sorted ),
	write( top_sort( CompYDir, Sorted ) ), nl,
	OutDag = rpdag(CompYPairs,CompYDir,CompYUnd).

in_pairs_add_arc( InPairs, A, B, ABPairs ) :-
	once( select(A-Aninf,InPairs,NoAPairs) ),
	once( select(B-Bninf,NoAPairs,NoBPairs) ),
	Aninf = (PaA/Pa,ChA/Ca,NeA/Na),
	Bninf = (PaB/Pb,ChB/Cb,NeB/Nb),
	% Add A to the parents of B
	ord_add_element( PaB, A, NwPaB ), NwPb is Pb + 1,
	% Add B to the children of A
	ord_add_element( ChA, B, NwChA ), NwCa is Ca + 1,
	NwAninf = (PaA/Pa,NwChA/NwCa,NeA/Na),
	NwBninf = (NwPaB/NwPb,ChB/Cb,NeB/Nb),
	% check that the following works as expected, ord_del_element/3 doesnt
	ord_add_element( NoBPairs, A-NwAninf, WtAPairs ),
	ord_add_element( WtAPairs, B-NwBninf, ABPairs ).
	
complete_fwd( [], Pairs, Dir, Und, Pairs, Dir, Und ).
complete_fwd( [Y|T], InPairs, InDir, InUnd, CompPairs, CompDir, CompUnd ) :-
	pick_head_undirected( [Y], InUnd, NwArcs/[], NxUnd ),
	% add_edges( InDir, NwArcs, NxDir ),
	direct_arcs( NwArcs, fwd, InPairs, InDir, NxPairs, NxDir, New ),
	% append( New, T, NxToComplete ),
	complete( New, NxPairs, NxDir, NxUnd, MdPairs, MdDir, MdUnd ),
	complete_fwd( T, MdPairs, MdDir, MdUnd, CompPairs, CompDir, CompUnd ).
% complete([3],[1-([]/0,[3]/1,[2]/1),2-([]/0,[]/0,[1,3]/2),3-([1,4]/2,[]/0,[2]/1),4-([]/0,[3]/1,[]/0)],[1-[3],2-[],3-[],4-[3]],[1-[2],2-[1,3],3-[2],4-[]],P,A,B)
% incorporate cycle detection here, and remove from other places (all?)
complete( [], Pairs, Dir, Und, Pairs, Dir, Und ).
complete( [Y|T], InPairs, InDir, InUnd, CompPairs, CompDir, CompUnd ) :-
	pick_head_undirected( [Y], InUnd, NwArcs/[], NxUnd ),
	% add_edges( InDir, NwArcs, NxDir ),
	direct_arcs( NwArcs, InPairs, InDir, NxPairs, NxDir, New ),
	append( New, T, NxToComplete ),
	complete( NxToComplete, NxPairs, NxDir, NxUnd, CompPairs, CompDir, CompUnd ).

pick_head_undirected( [], Und, Arcs/Arcs, Und ).
pick_head_undirected( [H|T], InUnd, Arcs/TArcs, CompUnd ) :-
	( select(H-HArcs,InUnd,MdUnd) ->
		directed_from_head_switches( HArcs, H, MdUnd, Arcs, NxUnd, MdArcs )
		;
		write( some_kind_of_warning(H) ), nl, abort,
		NxUnd = InUnd,
		MdArcs = Arcs
	),
	pick_head_undirected( T, NxUnd, MdArcs/TArcs, NoHCompUnd ),
	ord_add_element( NoHCompUnd, H-[], CompUnd ).

directed_from_head_switches( [], _From, Und, Arcs, Und, Arcs ).
directed_from_head_switches( [H|T], From, InUnd, [From-H|TArcs], Und, FArcs ) :-
	del_edges( InUnd, [H-From], NxUnd ),
	directed_from_head_switches( T, From, NxUnd, TArcs, Und, FArcs ).
% A.arc(x,y) ends

direct_arcs( [], Pairs, Dir, Pairs, Dir, [] ).
% direct_arcs( [A-B|T], InPairs, InDir, OutPairs, OutDir, Bs ) :-
	% add_edges( InDir, [B-A], NxDir ),
	% direct_arc( B, A, InPairs, NxPairs ),
	% direct_arcs( T, NxPairs, NxDir, OutPairs, OutDir, Bs ).
direct_arcs( [A-B|T], InPairs, InDir, OutPairs, OutDir, [B|Bs] ) :-
	add_edges( InDir, [A-B], NxDir ),
	direct_arc( A, B, InPairs, NxPairs ),
	direct_arcs( T, NxPairs, NxDir, OutPairs, OutDir, Bs ).

direct_arc( A, B, InPairs, OutPairs ) :-
	once( select(A-Aninf,InPairs,NoAPairs) ),
	once( select(B-Bninf,NoAPairs,NoBPairs) ),
	Aninf = (PaA/Pa,ChA/Ca,NeA/Na),
	Bninf = (PaB/Pb,ChB/Cb,NeB/Nb),
	% add A to parents of B
	ord_add_element( PaB, A, NwPaB ), NwPb is Pb + 1,
	% add B to children of A
	ord_add_element( ChA, B, NwChA ), NwCa is Ca + 1,
	% remove A from neighbours of B
	ord_del_element( NeB, A, NwNeB ), NwNb is Nb - 1,
	% remove B from neighbours of A
	ord_del_element( NeA, B, NwNeA ), NwNa is Na - 1,
	NwAninf = (PaA/Pa,NwChA/NwCa,NwNeA/NwNa),
	NwBninf = (NwPaB/NwPb,ChB/Cb,NwNeB/NwNb),
	% check that the following works as expected, ord_del_element/3 doesnt
	ord_add_element( NoBPairs, A-NwAninf, WtAPairs ),
	ord_add_element( WtAPairs, B-NwBninf, OutPairs ).
