:- ensure_loaded( library(lists) ).			% memberchk/2,
 										% working_directory/2.
 										% append/3.
:- ensure_loaded( '../../src/init_lib' ).		%.
:- requires( directory_files_regular/2 ).
:- requires( break_nth/4 ).
:- requires( contains/2 ).
:- requires( file_stems_of_files_with_this_ext/3 ).

bst_mst_table :-
	findall( Prior, prior_dir(Prior), Priors ),
	priors_matrix( Priors, PriorsMtx ),
	open( 'BstMstTable.txt', write, Out ),
	write( Out, '            id            ed            ne                  ks                ck' ), nl( Out ),
	TrueBNs = [as,al,hf,sc,p1,p2],
	write_priors_matrix( TrueBNs, PriorsMtx, Out ), 
	close( Out ).

write_priors_matrix( [], _Mtx, _Out ).
write_priors_matrix( [HBn|TBns], PrMtx, Out ) :-
	all_bn_entries( PrMtx, HBn, Bn500Ents, Bn1KEnts ),
	true_bn_name( HBn, BnName ),
	write( Out, BnName ), nl( Out ),
	write( Out, 'datapoints: 500' ), nl( Out ),
	write( Out, 'Prior' ),
	write_bn_entries( Bn500Ents, Out ),
	nl( Out ),
	write( Out, 'datapoints: 1000' ), nl( Out ),
	% write_loss_spaces( 9, Out ),
	write( Out, 'Prior' ),
	write_bn_entries( Bn1KEnts, Out ),
	nl( Out ),
	write_priors_matrix( TBns, PrMtx, Out ).

all_bn_entries( [], _HBn, [], [] ).
all_bn_entries( [HPr/HPrEnt|T], HBn, [H500|TBn500Ents], [H1K|TBn1KEnts] ) :-
	memberchk( HBn+row(A,B,C,D,E,F), HPrEnt ),
	H500 = HPr/row(A,B,C),
	H1K  = HPr/row(D,E,F),
	all_bn_entries( T, HBn, TBn500Ents, TBn1KEnts ).

write_bn_entries( [], _Out ).
write_bn_entries( [HPr/row(A,B,C)|T], Out ) :-
	atom_concat( prior, PriorId, HPr ),
	write( Out, '  ' ), write( Out, PriorId ), write( Out, 'b: ' ),
	% write_bn_losses( A, B, C ), 
	loss_in_lists( id, A, B, C, IdLoBst, IdLoMst ),
	loss_in_lists( ed, A, B, C, EdLoBst, EdLoMst ),
	loss_in_lists( ne, A, B, C, NeLoBst, NeLoMst ),
	loss_in_lists( ks, A, B, C, KsLoBst, KsLoMst ),
	loss_in_lists( ck, A, B, C, CkLoBst, CkLoMst ),
	write_loss_numbers( IdLoBst, 3, Out ),
	write_loss_numbers( EdLoBst, 3, Out ),
	write_loss_numbers( NeLoBst, 5, Out ),
	write_loss_numbers( KsLoBst, 5, Out ),
	write_loss_numbers( CkLoBst, 5, Out ),
	nl( Out ), write( Out, '         m: ' ),
	write_loss_numbers( IdLoMst, 3, Out ),
	write_loss_numbers( EdLoMst, 3, Out ),
	write_loss_numbers( NeLoMst, 5, Out ),
	write_loss_numbers( KsLoMst, 5, Out ),
	write_loss_numbers( CkLoMst, 5, Out ),
	nl( Out ), write( Out, '     ' ),
	write_bn_entries( T, Out ).

%here
loss_in_lists( Name, Ab/Am, Bb/Bm, Cb/Cm, LoBst, LoMst ) :-
	loss_memberchk( Name, Ab, AbArg ),
	loss_memberchk( Name, Am, AmArg ),
	loss_memberchk( Name, Bb, BbArg ),
	loss_memberchk( Name, Bm, BmArg ),
	loss_memberchk( Name, Cb, CbArg ),
	loss_memberchk( Name, Cm, CmArg ),
	LoBst = [AbArg,BbArg,CbArg],
	LoMst = [AmArg,BmArg,CmArg].

loss_memberchk( Name, List, Arg ) :-
	( List == na -> 
		Arg = na 
		;
		Item =.. [Name,Arg], 
		( memberchk(Item,List) ->
			true
			;
			Arg = na
		)
	).

%here
write_loss_numbers( [], _Lgth, _Out ).
write_loss_numbers( [H|T], Lgth, Out ) :-
	( H == na ->
		RedLgth is Lgth - 1,
		write_loss_spaces(RedLgth,Out),
		write( Out, '_' )
		;
		number_codes( H, HCsPrv ),
		( HCsPrv = [0'0|HCsPrv2] -> 
			( HCsPrv2 == [] ->
				HCs = HCsPrv 
				;
				HCs = HCsPrv2
			)
			;
			HCs = HCsPrv
		),
		length( HCs, HLgth ),
		( HLgth > Lgth ->
			break_nth( Lgth, HCs, HoutCs, _ )
			;
			HCs = HoutCs,
			( HLgth < Lgth -> 
				Diff is Lgth - HLgth,
				write_loss_spaces(Diff,Out)
				;
				true
			)
		),
		atom_codes( Hout, HoutCs ),
		write( Out, Hout )
	),
	( T == [] -> write(Out,'  :') ; write(Out,',') ),
	write_loss_numbers( T, Lgth, Out ).

priors_matrix( [], [] ).
priors_matrix( [H|T], [HRow|TRows] ) :-
	atom_concat( H, '/losses', LossesDir ),
	working_directory( Top, LossesDir ),
	directory_files_regular( '.', All ),
	TrueBNs = [as,al,hf,sc,p1,p2],
	file_stems_of_files_with_this_ext( All, '_bst_mst.pl', BstMst ),
	atom_codes_list( BstMst, BstMstCs ),
	prior_row( TrueBNs, BstMstCs, Row ),
	HRow = H/Row,
	working_directory( _, Top ),
	priors_matrix( T, TRows ).

prior_row( [], _BstMst, [] ).
prior_row( [HBn|TBns], BstMst, [HBn+HEnt|TRows] ) :-
	atom_codes( HBn, HBnCsPrv ),
	append( "cm_t", HBnCsPrv, HBnCs ),
	bst_mst_prior_entry( BstMst, HBnCs, "d500", "s103", A, RemBstMstA ),
	bst_mst_prior_entry( RemBstMstA, HBnCs, "d500", "s451", B, RemBstMstB ),
	bst_mst_prior_entry( RemBstMstB, HBnCs, "d500", "s772", C, RemBstMstC ),
	bst_mst_prior_entry( RemBstMstC, HBnCs, "d1K", "s103", D, RemBstMstD ),
	bst_mst_prior_entry( RemBstMstD, HBnCs, "d1K", "s451", E, RemBstMstE ),
	bst_mst_prior_entry( RemBstMstE, HBnCs, "d1K", "s772", F, RemBstMstF ),
	HEnt = row(A,B,C,D,E,F),
	prior_row( TBns, RemBstMstF, TRows ).

bst_mst_prior_entry( [], _TrBn, _Data, _Seed, na/na, [] ).
bst_mst_prior_entry( [H|T], TrBn, Data, Seed, Ent, Rem ) :-
	contains( H, TrBn ), 
	contains( H, Data ), 
	contains( H, Seed ),
	!,
	atom_codes( HAtm, H ),
	atom_concat( HAtm, '_bst_mst.pl', File ),
	open( File, read, In ),
	read( In, bst_lgllhood_losses(BstLosses,_,_,_) ),
	read( In, mst_visited_losses(MstLosses,_,_,_) ),
	Ent = BstLosses/MstLosses,
	close( In ),
	Rem = T.
bst_mst_prior_entry( [H|T], TrBn, Data, Seed, Ent, [H|Rem] ) :-
	bst_mst_prior_entry( T, TrBn, Data, Seed, Ent, Rem ).

prior_dir( prior01 ).
prior_dir( prior02 ).
prior_dir( prior03 ).
prior_dir( prior04 ).
prior_dir( prior05 ).
prior_dir( prior06 ).
prior_dir( prior07 ).
prior_dir( prior08 ).
prior_dir( prior09 ).

true_bn_name( as, 'Asia      ' ).
true_bn_name( al, 'Alarm     ' ).
true_bn_name( hf, 'Hailfinder' ).
true_bn_name( sc, 'Insurance ' ).
true_bn_name( p1, 'Parity1   ' ).
true_bn_name( p2, 'Parity2   ' ).

atom_codes_list( [], [] ).
atom_codes_list( [H|T], [HCs|TCs] ) :-
	atom_codes( H, HCs ),
	atom_codes_list( T, TCs ).

write_loss_spaces( 0, _Out ) :- !.
write_loss_spaces( N, Out ) :-
	write( Out, ' ' ),
	NxN is N - 1,
	write_loss_spaces( NxN, Out ).
