
:- ensure_loaded( library(lists) ).				% memberchk/2, append/3.
:- ensure_loaded( library(system) ).				% file_exists/1.
:- ensure_loaded( '../../src/init_lib' ).

:- requires( opts_have_help/1 ).
:- requires( select_compound_w_gunzip/4 ).
:- requires( gzip_files/1 ).
:- requires( sys_info/1 ).

loss_gsplay( Opts ) :-
	opts_have_help( Opts ),
	!,
	write( 'loss_gsplay       create postscript plot outputs for a number of loss files' ), nl, nl,
	write( ' Options.' ), nl,
	write( '       length:    length of loss files' ), nl, 
	write( '                  any non compound argument is taken to be a loss file. ' ), nl, nl.


loss_gsplay( Opts ) :-
	( memberchk( length(Lgt), Opts ) ->
		( atom(Lgt) ->
			atom_concat( Lgt, ' ', LgtSpc )
			;
			number_codes( Lgt, LgtCs ),
			append( LgtCs, " ", LgtSpcCs ),
			atom_codes( LgtSpc, LgtSpcCs )
		)
		;
		true
	),
	select_compound_w_gunzip( Opts, _Cmp, Files, Gz ),
	loss_gsplay_files( Files, LgtSpc ),
	gzip_files( Gz ).

loss_gsplay_files( [], _Lgt ).
loss_gsplay_files( [H|T], Lgt ) :-
	( min_max_file( H, YMin, YMax ) ->
		number_codes( YMin, YMinCs ),
		number_codes( YMax, YMaxCs ),
		atom_codes( YMinAtm, YMinCs ),
		atom_codes( YMaxAtm, YMaxCs ),
		atom_concat( ' ypad=min=', YMinAtm, Ymin ),
		atom_concat( ' ypad=max=', YMaxAtm, Ymax ),
		atom_concat( Ymax, ' ', YmaxSpc ),
		atom_concat( Ymin, YmaxSpc, YMinMax )
		;
		YMinMax = ' '
	),
	( var(Lgt) -> 
		XMinMax = ' '
		;
		atom_concat( ' xpad=max=', Lgt, XMaxAtm ),
		atom_concat( ' xpad=min=0 ', XMaxAtm, XMinMax )
	),
	atom_concat( 'upsh gsplay term=\'postscript color\' output=\'', H, Pfx1 ),
	atom_concat( Pfx1, '.ps\' ', Pfx2 ),
	atom_concat( Pfx2, XMinMax, Pfx3 ),
	atom_concat( Pfx3, YMinMax, Pfx4 ),
	atom_concat( Pfx4, H, Upsh ),
	write( upsh(Upsh) ), nl,
	sys_info( Upsh ),
	loss_gsplay_files( T, Lgt ).

min_max_file( File, Min, Max ) :-
	atom_codes( File, FileCs ),
	append( Stem, [0'.,_,A,B|_Rest], FileCs ),
	!,
	append( Stem, [0'.,0'm,A,B], MFileCs ),
	atom_codes( MFile, MFileCs ),
	file_exists( MFile ),
	open( MFile, read, MStream ),
	read( MStream, min_max(Min,Max) ),
	close( MStream ).
