:- ensure_loaded( library(ordsets) ).		% ord_add_element/3.
:- ensure_loaded( '../../src/init_lib' ).
:- ensure_loaded( library(read_term_frequencies) ).
:- ensure_loaded( library(gnucomfile_new) ).
:- ensure_loaded( library(kv_decompose) ).			% 3.
:- ensure_loaded( library(kv_compose) ).			% 3.

sample_files_comparative_plot( Opts ) :-
	Files = Opts, % for now.
	read_frqs_from_files( Files, AbsFrqs ),
	keysort_all( AbsFrqs, SrtFrqs ),
	to_relative_frqs( SrtFrqs, Frqs ),
	all_terms_in_frqs( Frqs, [], AllTerms ),
	complete_frqs_with_unseen_from( Frqs, AllTerms, CompFrqs ),
	% check_unseen( CompFrqs, 1 ),
	% its_ordered( CompFrqs ),
	amortisise_to_first_s_frqs( CompFrqs, AmoFrqs ),
	store_completed_frqs( AmoFrqs, Files, Stores ),
	multi_plot( Stores ).

read_frqs_from_files( [], [] ).
read_frqs_from_files( [H|T], [HFs|TFs] ) :-
	read_term_frequencies( H, HFs ),
	read_frqs_from_files( T, TFs ).

all_terms_in_frqs( [], All, All ).
all_terms_in_frqs( [H|T], Acc, All ) :-
	add_terms_of_frq( H, Acc, Mid ),
	all_terms_in_frqs( T, Mid, All ).

add_terms_of_frq( [], Add, Add ).
add_terms_of_frq( [K-_|T], Acc, Add ) :-
	% kvs_add_an_element( Acc, K, Mid ),
	ord_add_element( Acc, K, Mid ),
	add_terms_of_frq( T, Mid, Add ).

complete_frqs_with_unseen_from( [], _All, [] ).
complete_frqs_with_unseen_from( [H|T], All, [CompH|CompT] ) :-
	complete_frq_with( H, All, CompH ),
	complete_frqs_with_unseen_from( T, All, CompT ).

complete_frq_with( [], All, Comp ) :-
	complete_frq_with_rem( All, Comp ).
complete_frq_with( [Kh-Vh|T], [Ka|Ta], Comp ) :-
	( Ka @< Kh ->
		Comp = [Ka-0|TComp], NxT = [Kh-Vh|T]
		;
		Comp = [Kh-Vh|TComp], NxT = T
	),
	complete_frq_with( NxT, Ta, TComp ).

complete_frq_with_rem( [], [] ).
complete_frq_with_rem( [H|T], [H-0|TComp] ) :-
	complete_frq_with_rem( T, TComp ).

store_completed_frqs( [], [], [] ).
store_completed_frqs( [H|T], [F|Fs], [S|Ss] ) :-
	atom_concat( Stem, '.pl', F ), 
	atom_concat( Stem, '_frq_cmp', S ),
	open( S, write, Out ),
	write_complete_frq( H, Out ),
	close( Out ),
	store_completed_frqs( T, Fs, Ss ).

write_complete_frq( [], _S ).
write_complete_frq( [K-_V|T], S ) :-
	write( S, K ),  nl( S ),
	write_complete_frq( T, S ).

multi_plot( Stores ) :-
	gnucomfile( gfile, [], Stores ).

to_relative_frqs( [], [] ).
to_relative_frqs( [H|T], [F|Fs] ) :-
	kv_add_vs( H, 0, Sum ),
	to_relative_frq( H, Sum, F ),
	to_relative_frqs( T, Fs ).

to_relative_frq( [], _Sum, [] ).
to_relative_frq( [K-V|T], Sum, [K-RelV|RelT] ) :-
	RelV is V / Sum,
	to_relative_frq( T, Sum, RelT ).
	
kv_add_vs( [], Sum, Sum ).
kv_add_vs( [_K-V|T], Acc, Sum ) :-
	NxAcc is Acc + V,
	kv_add_vs( T, NxAcc, Sum ).

amortisise_to_first_s_frqs( [H|T], [HA|TAs] ) :-
	kv_decompose( H, HKs, HVs ),
	kv_compose( HVs, HKs, TrH ),
	keysort( TrH, HRevA ),
	reverse( HRevA, HA ),
	amortisise_frqs_to_this( T, HA, TAs ).

amortisise_frqs_to_this( [], _This, [] ).
amortisise_frqs_to_this( [H|T], This, [HA|TAs] ) :-
	amortisises_frqs( This, H, HA ),
	amortisise_frqs_to_this( T, This, TAs ).

amortisises_frqs( [], Frqs, [] ) :-
	( Frqs = [] -> true
		; 
		write( user_error, 'Non empty residual in amortisises_frqs/3, see log file.' ), 
		nl( user_error ),
		open( 'log.pl', append, Log ),
		write( Log, frqs(Frqs)), nl( Log ),
		close( Log )
	).
amortisises_frqs( [_-Term|T], Frqs, Amor ) :-
	( sel_element(Frqs,Term,TrmFq,Red) ->
		Amor = [TrmFq-Term|TAmor]
		;
		write( failed(term(Term),tfg(TrmFq)) ), nl,
		abort,
		Amor = TAmor,
		Red = Frqs,
		write( user_error, 'Couldnt not pick term: ' ),
		write( user_error, Term ), nl( user )
	),
	% its_ordered_1( Red ),
	amortisises_frqs( T, Red, TAmor ).

% sel_del_element( [], Term-TrmFq, Red) :-
sel_element( [K-V|T], Sel, SelV, Red) :-
	K @< Sel,
	!,
	Red = [K-V|TRed],
	sel_element( T, Sel, SelV, TRed ).
sel_element( [K-V|T], Sel, SelV, Red) :-
	Sel == K,
	SelV = V,
	Red = T.
% sel_element( [K-V|T], Sel, SelV, Red) :- 
	% write( k(K) ), nl,
	% write( v(V) ), nl,
	% write( sel(Sel) ), nl,
	% write( sv(SelV) ), nl,
	% abort.

kvs_add_an_element( [Hk-Hv|T], K-V, [Hk-Hv|TAll] ) :-
	Hk @< K,
	!,
	kvs_add_an_element( T, K-V, TAll ).
kvs_add_an_element( List, K-V, [K-V|List] ).

keysort_all( [], [] ).
keysort_all( [H|T], [SrtH|TSs] ) :-
	keysort( H, SrtH ),
	keysort_all( T, TSs ).

check_unseen( [F,S|T], N ) :- 
	check_those( F, S ),
	NxN is N + 1,
	check_unseen( [S|T], NxN ).
check_unseen( [_F], _N ).

check_those( [], _Lest2 ).
check_those( [Trm1-_|T1], [Trm1-_|T2] ) :-
	check_those( T1, T2 ).
check_those( [Trm1-_|T1], [Trm2-_|T2] ) :-
	write( clash(t1,Trm1) ), nl,
	write( clash(t2,Trm2) ), nl,
	abort,
	check_those( T1, T2 ).

its_ordered( [] ).
its_ordered( [H|T] ) :-
	its_ordered_1( H ), 
	its_ordered( T ).

its_ordered_1( [_F] ).
its_ordered_1( [F-_FV,S-SV|T] ) :-
	F @< S,
	!,
	its_ordered_1( [S-SV|T] ).
its_ordered_1( [F-_FV,S-_SV|_T] ) :-
	write( ahaf(F) ), nl,
	write( ahas(S) ), nl, abort.
