:- use_module(library(ordsets)).
:- use_module(library(lists)).

:- op( 100, xfx, '<-' ).

% Build a BN with an accumulator

/*---------------------------------------------------------------------

InVars - variables in InBN
InDeps - dependencies as an ordered list, e.g. [a-[],b-[c],c-[b]]
InBn - BN so far as a list of directed edges

---------------------------------------------------------------------*/

% debugging k([],_AllVars,Deps,In,In) :- write(Deps).
k([],_,_,In,In).
k([RV|RVs],InVars,InDeps,InBN,OutBN) :-
	edges(InVars,RV,InDeps,InBN,MidDeps,MidBN),
	k(RVs,[RV|InVars],MidDeps,MidBN,OutBN).
        

edges([],_RV,Deps,BN,Deps,BN).                       % finished
edges([H|T],RV,InDeps,TailBN,OutDeps,BN) :-          % connect
	add_child(RV,H,InDeps,MidDeps),
        edges(T,RV,MidDeps,[RV<-H|TailBN],OutDeps,BN).
edges([_H|T],RV,InDeps,TailBN,OutDeps,BN) :-          % don't connect
        edges(T,RV,InDeps,TailBN,OutDeps,BN).

% Assumes that desired independencies are given as follows:
% independent(a,b)
% independent(b,d)
% etc

% indep(b,c).
%% make sure indep( Nd1, Nd2 ) -> Nd1 @< Nd2.

/*---------------------------------------------------------------------
  Predicate: add_child(+Child,+Parent,+Deps,-NewDeps)
       Args: Child is a node being added to BN via a proposed link from Parent
   Succeeds: 0 or 1
   Comments: BN is assumed to respect all independence assumptions
             Is true if adding a Child<-Parent link to BN does not transgress
	     any independence assumptions
	     Deps is of the form [a-b,b-c,..] for each node the other
	     nodes it depends on is given
---------------------------------------------------------------------*/

add_child(Child,Parent,Deps,NewDeps) :-
	\+ indep(Parent,Child),
	%% \+ independent(Child,Parent),
	\+ (
	     (
	       independent(Child,Var),
	       are_dependent(Var,Parent,Deps)
	    )
	   ),
	get_others(Deps,Parent,[],Others),
	add_dependencies([Parent|Others],Child,Deps,NewDeps).

add_dependencies([],_Child,NewDeps,NewDeps).
add_dependencies([H|T],Child,Deps,NewDeps) :-
	add_dependency(H,Child,Deps,MidDeps),
	add_dependencies(T,Child,MidDeps,NewDeps).

add_dependency(X,Y,In,Out) :-
	X @< Y,
	!,
	ord_add_element(In,X-Y,Out).
add_dependency(X,Y,In,Out) :-
	ord_add_element(In,Y-X,Out).

get_others([],_Parent,Others,Others).
get_others([X-Parent|T],Parent,In,Out) :-
	!,
	get_others(T,Parent,[X|In],Out).
get_others([Parent-X|T],Parent,In,Out) :-
	!,
	get_others(T,Parent,[X|In],Out).
get_others([_|T],Parent,In,Out) :-
	get_others(T,Parent,In,Out).

are_dependent(X,Y,Deps) :-
	X @< Y,
	!,
	ord_member(X-Y,Deps).
are_dependent(X,Y,Deps) :-
	ord_member(Y-X,Deps).



independent(X,Y) :-
	indep(X,Y).
independent(X,Y) :-
	indep(Y,X).
