:- module( introduces, 
			[
				edge_introduces_n_inbreedings/4,
				edge_introduces_n_promiscuities/4,
				edge_introduces_generation/4,
				max_length_path_from/3 ] ).

:- compile( sicstus_path ).

:- ensure_loaded( library(lists) ).		% member/2, memberchk/2.
:- ensure_loaded( library(ugraphs) ).		% reachable/3, min_path/5.
:- ensure_loaded( library(ordsets) ).		% ord_intersect/2
 
% compile( introduces ), save_modules( [introduces], introduces ), halt.

edge_introduces_n_inbreedings( Person, Parent, Ped, N ) :-
	Ped = ped(AncG,DscG,Prtns), 
	reachable( Person, DscG, PsDscs ),
	findall( PromDsc, 
			(
				member(PromDsc,PsDscs),
				partner_of(PromDsc,Prtns,DscPartner),
				min_path_pl(DscPartner,Parent,AncG,_MnPath,_MinCost ),
				\+ incestious_pair( PromDsc, DscPartner, AncG )
			), PromDscs ),
	length( PromDscs, N ).

edge_introduces_n_promiscuities( _Person, Parent, DscG, N) :-
	memberchk( Parent-OthChildren, DscG ), 
	% this is quite crude. maybe in the future we can 
	% select both parents in one step.
	length( OthChildren, N ).

edge_introduces_generation( DscL, Parent, AncG, L ) :-
	max_length_path_from( Parent, AncG, AncL ),
	L is DscL + AncL + 1. 

% put in different files

partner_of( Person, Partners, Partner ) :-
	memberchk( Person-PersonPartners, Partners ),
	member( Partner, PersonPartners ).

incestious_pair( PersonA, PersonB, AncG ) :-
	reachable( PersonA, AncG, ReachedByA ),
	reachable( PersonB, AncG, ReachedByB ),
	ord_intersect( ReachedByA, ReachedByB ).

max_length_path_from( Person, Graph, Max ) :-
	findall( Length, 
			( 
			  path_pl(Person,Graph,Path), 
			  length(Path,Length)
			),
			  		Lengths ),
	max_list_w_empty( Lengths, Max ).

max_list_w_empty( List, Max ) :-
	( List = [] -> Max is 0 ; max_list( List, Max ) ).
