:- ensure_loaded( library(lists) ).	% append/3, select/3.

:- requires( band_bn/3 ).
:- requires( kv_to_ord_k_v/3 ).

%  bn_to_dot_dag( [1-[],2-[1],3-[2,5],4-[],5-[4],6-[4],7-[3],8-[3,6]], [], tara ).

bn_to_dot_dag( Bn, Against, File ) :-
	bn_to_dot_dag( Bn, Against, File, [] ).

% Opts in {colour(colour/bnw),node_style(filled/empty)} type(digraph/graph)...
%
bn_to_dot_dag( Bn, Against, File, Opts ) :-
	bn_to_dot_dag_defaults( Defs ),
	append( Opts, Defs, AllOs ),
	memberchk( node_style(NdSt), AllOs ),
	memberchk( colour(Clr), AllOs ),
	memberchk( type(Type), AllOs ),
	memberchk( collapse_bi(Collapse), AllOs ),
	memberchk( title(Title), AllOs ),
	open( File, append, FStream ),
	current_output( Old ),
	set_output( FStream ),
     memberchk( def_edge_colour(EdgeClr), AllOs ),
	dot_preample( NdSt, Type, EdgeClr ),
	band_bn( Bn, Collapse, Bands ),
	kv_to_ord_k_v( Against, AgsCh, AgsPa ),
	induce_parentless( AgsPa, AgsCh, [], _Pless, Induced ), 
	add_induced_parentless( Induced, Against, CompAgs ),
     ( Collapse == true -> 
          clear_bidirectional( CompAgs, NoBiAgainst )
          ;
          CompAgs = NoBiAgainst
     ),
	bn_bands_to_dot_dag_output( Bands, Title, NoBiAgainst, Type, Clr, AllOs ),
	set_output( Old ),
	close( FStream ).

bn_to_dot_dag_defaults( Defs ) :- 
     Defs = [colour(colour),node_style(filled),
             type(digraph), nodes_colours([]), 
             def_edge_colour(black),collapse_bi(true),
             nodes_labels([]),edges_colours([]),title(false)].

bn_bands_to_dot_dag_output( Bands, Title, Against, Type, Clr, Opts ) :-
	memberchk( nodes_colours(NdClrs), Opts ),
	memberchk( nodes_labels(NdLbls), Opts ),
	memberchk( edges_colours(EdgeClrs), Opts ),
	bands_to_zones( Bands, Zones ),
     dot_title( Title, Zones ),
	dot_zones_to_colours( Zones, Clr, NdClrs, NdLbls ),
     graph_type_to_edge_atom( Type, EdgeAtm ),
	dot_edges( Bands, Against, EdgeAtm, EdgeClrs ),
	write( '}' ), nl.

bands_to_zones( [Pless,Adjusted,Cless], [Pless,AdjZone,ClessZone] ) :-
	bands_to_zones1( Adjusted, AdjZone ),
	bands_to_zones1( Cless, ClessZone ).

bands_to_zones1( [], [] ).
bands_to_zones1( [PrvH-_HPs|T], [H|M] ) :-
	( PrvH = dbl(H) -> true ; H = PrvH ),
	bands_to_zones1( T, M ).

dot_title( false, _Zones ) :- !.
dot_title( Title, Zones ) :-
     findall( X, (member(Z,Zones),member(0-X,Z)), Xs ),
     ( Xs == [] ->
          TitleNodeId is 0
          ;
          TitleNodeId = 'title'
     ),
     tab( 2 ), write( TitleNodeId ), write( ' ' ),
     Opts = [shape(invhouse),label(Title)],
     print_graphviz_entry_opts( Opts ).

dot_zones_to_colours( [PlessZ,AdjZ,ClessZ], Clr, NdClrs, NdLbls ) :-
	( Clr == bnw -> 
		Red = black, Orange = black, Green = black
		;
		Red = red, Orange = orange, Green = green
	),
	dot_zone_to_colour( PlessZ, Red, NdClrs, NdLbls ),
	dot_zone_to_colour( AdjZ, Orange, NdClrs, NdLbls ),
	dot_zone_to_colour( ClessZ, Green, NdClrs, NdLbls ).

dot_zone_to_colour( [], _Colour, _NdColours, _NdLbls ).
dot_zone_to_colour( [H|T], Colour, NdColours, NdLbls ) :-
     ( memberchk(H-NdClr,NdColours) -> true; NdClr = Colour ),
	tab( 2 ), write( H ), tab( 1 ),
     ( memberchk(H-NdLbl,NdLbls) -> 
         write( '[' ), 
         write( color='"' ), write( NdClr ), write( '"' ),
         write( ', label'='"' ), write( NdLbl ), write( '"' ),
         write( ']' )
         ;
         write( '[' ), 
         write( color='"' ), write( NdClr ), write( '"' ),
         write( ']' )
     ),
	nl,
	dot_zone_to_colour( T, Colour, NdColours, NdLbls ).

dot_edges( [Pless,Adj,Cless], Against, EdgeAtm, Clrs ) :-
	kv_newval_empty_list( Pless, KvPless ),
	dot_edges_one( KvPless, Against, EdgeAtm, Clrs ),
	dot_edges_one( Adj, Against, EdgeAtm, Clrs ),
	dot_edges_one( Cless, Against, EdgeAtm, Clrs ).

dot_edges_one( [], _Against, _EdgeAtm, _Clrs ).
dot_edges_one( [Node-Parents|T], Against, EdgeAtm, Clrs ) :-
	% HERE select( Node-AgainstPrs, Against, _Ragain ),
	memberchk( Node-AgainstPrs, Against ),
	dot_node_parents_edges( Parents, AgainstPrs, EdgeAtm, Node, Clrs ),
	dot_edges_one( T, Against, EdgeAtm, Clrs ).

% 11/2010:  CLrs expanded to any edge option. format is  X-Y-List
% List = [v(t)|_]  -> v=t
dot_node_parents_edges( [], Against, _EdgeAtm, Node, Clrs ) :-
	dot_node_parents_edges_vanished( Against, Node, Clrs ).
dot_node_parents_edges( [PrvH|T], Against, EdgeAtm, Node, Clrs ) :-
	% ( PrvH = dbl(H) -> Dbl = true ; H = PrvH, Dbl = false  ),
	( PrvH = dbl(H) ->
          % throw( obsolete_double_notation(H,Node) )
          true
          ;
          H  = PrvH
     ),
	tab( 2 ), write( H ), write( EdgeAtm ), write( Node ),
	( select(PrvH,Against,RemAgainst) ->
          true
          ;
          throw( against_no_longer_supported__load_to_edges_opts(H) )
     ),
     ( select(H-Node-EOpts,Clrs,RemClrs) -> 
          (PrvH = dbl(_) -> add_opt(dir,EOpts,none,DblOpts); DblOpts = EOpts)
          ; 
          RemClrs = Clrs,
          DblOpts = []
     ),
     print_graphviz_entry_opts( DblOpts ),
     /*
	( select(H,Against,RemAgainst) ->
		( Dbl == true ->
               ( memberchk(H-Node-Colour,Clrs) -> true; Colour = black ),
			write( ' [color="' ), 
               write( Colour ), write( '", dir=none]' ), nl
			; 
               % added the following 2 lines in 11/2010
               ( memberchk(H-Node-Colour,Clrs) -> true; Colour = black ),
               write( '[colour="' ), write( Colour ), write( '"]' ),
               nl
		)
		;
		RemAgainst = Against,
		( Dbl == true ->
               ( memberchk(H-Node-Colour,Clrs) ->
			          write( ' [color="' ),
                         write( Colour ), write( '",' )
                         ; 
                           % leave it to the default so graph-wide edge colour
                           % is applied
                         write( '[' )
               ),
               write( ' dir=none]' ), nl
			;
               ( memberchk(H-Node-Colour,Clrs) -> true; Colour = magenta),
			write( ' [color="' ), write( Colour ), write( '"]' ), nl
		)
	),
     */
	dot_node_parents_edges( T, RemAgainst, EdgeAtm, Node, RemClrs ).

dot_node_parents_edges_vanished( [], _Node, _Clrs ).
dot_node_parents_edges_vanished( [H|T], Node, Clrs ) :-
	( H = dbl(_) ->
		true
		;
		tab( 2 ), write( H ), write( '->' ), write( Node ),
          ( memberchk(H-Node-Colour,Clrs) ->
               true
               ;
               Colour = yellow
          ),
		write( ' [color=' ),
          write( Colour ),
          write( ']' ), nl,
		dot_node_parents_edges_vanished( T, Node, Clrs )
	).

kv_newval_empty_list( [], [] ).
kv_newval_empty_list( [H|T], [H-[]|Te] ) :-
	kv_newval_empty_list( T, Te ).

graph_type_to_edge_atom( graph, '--' ).
graph_type_to_edge_atom( digraph, '->' ).

dot_preample( EdgStyle, Type, EdgeClr ) :-
     write( Type ),
	write( ' G {' ), nl,
	write( '  graph [URL="default.html", BGURL="smbluewhite_paper.gif "]' ), nl,
	write( '  node [style=' ), write( EdgStyle ),
	write( ']' ), nl, nl,
	write( '  edge [color=' ), 
     write( EdgeClr ),
     /*
	( Clr == bnw -> 
		write( black )
		;
		write( red )
	),
     */
	write( ']' ), nl.

print_graphviz_entry_opts( Opts ) :-
     write( ' [' ), 
     print_graphviz_entry_opts_rec( Opts ),
     write( ']' ), nl.

print_graphviz_entry_opts_rec( [] ).
print_graphviz_entry_opts_rec( [H|T] ) :-
     ( (  functor( H, Name, 1 ),
          arg( 1, H, Arg ),
          write( Name=Arg ),
          ( T \== [] -> write( ', ' ) ; true )
       ) ->    true
               ;
               throw( wrong_arity_in_option(H) )
     ),
     print_graphviz_entry_opts_rec( T ).

add_opt( Name, InOpts, Value, OutOpts ) :-
     functor( Term, Name, 1 ),
     ( memberchk(Term,InOpts) ->
          throw( trying_to_add_opt_with_existing_value(Name,InOpts) )
          ;
          true
     ),
     arg( 1, Term, Value ),
     OutOpts = [Term|InOpts].
